from typing import Union, List, Set

from platon_env.base.host import Host
from platon_env.base.service import Service
from platon_env.node import Node, NodeOpts
from platon_env.factory import chain_factory
from platon_env.genesis import Genesis
from platon_env.utils.executor import concurrent_executor


class Chain(Service):
    hosts: Set[Host] = set()
    init_nodes: Set[Node] = set()
    normal_nodes: Set[Node] = set()

    def __init__(self, nodes: List[Node] = None):
        """ 初始化chain对象
        """
        super().__init__()
        if not nodes:
            nodes = []
        for node in nodes:
            self.add_process(node)

    @property
    def nodes(self):
        return set.union(self.init_nodes, self.normal_nodes)

    def install(self,
                platon: str,
                network: str,
                genesis_file: str = None,
                static_nodes: List[str] = None,
                keystore: str = None,
                options: str = '',
                nodes: List[Node] = None,
                ):
        """ 部署链
        """
        nodes = nodes or self.nodes
        if network == 'private' and genesis_file:
            self.full_genesis_file(genesis_file, nodes)
            static_nodes = static_nodes or [node.enode for node in nodes]

        return concurrent_executor(nodes,
                                   'install',
                                   platon,
                                   network,
                                   genesis_file,
                                   static_nodes,
                                   keystore,
                                   options,
                                   )

    def uninstall(self, nodes: List[Node] = None):
        """ 清理链，会停止节点并删除节点文件
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'uninstall')

    def add_process(self, node: Node):
        """ 将进程添加到服务，进行统一管理
        """
        if self.processes.get(id(node)):
            raise Exception('The node already exists.')

        self.init_nodes.add(node) if node.is_init_node else self.normal_nodes.add(node)
        self.hosts.add(node.host)

        self.processes[id(node)] = node

    def status(self, nodes: List[Node] = None):
        """ 检查链运行状态
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'status')

    def init(self, nodes: List[Node] = None):
        """ 初始化链
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'init')

    def start(self,
              options: Union[str, NodeOpts] = '',
              nodes: List[Node] = None,
              ):
        """ 启动链
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'start', options)

    def restart(self, nodes: List[Node] = None):
        """ 重启链
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'restart')

    def stop(self, nodes: List[Node] = None):
        """ 停止链
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'stop')

    def upload_platon(self, platon_file, nodes: List[Node] = None):
        """ 使用缓存上传platon
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'upload_platon', platon_file)

    def upload_keystore(self, keystore_path, nodes: List[Node] = None):
        """ 使用缓存上传keystore并解压
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'upload_keystore', keystore_path)

    def set_static_nodes(self, enodes: List[str], nodes: List[Node] = None):
        """ 指定要连接的静态节点，可以指定多个
        """
        nodes = nodes or self.nodes
        return concurrent_executor(nodes, 'set_static_nodes', enodes)

    def full_genesis_file(self, genesis_file, nodes: List[Node] = None):
        """ 填充创世文件，目前仅填充初始验证节点
        """
        nodes = nodes or self.nodes
        init_node = [node for node in nodes if node.is_init_node]
        genesis = Genesis(genesis_file)
        if init_node:
            genesis.fill_init_nodes(init_node)

        if not genesis.init_node:
            raise ValueError('the genesis file init node is empty, and no init node in the nodes argument.')

        genesis.save(genesis_file)

    @staticmethod
    def from_file(file):
        """ 通过chain肖像文件, 生成chain对象
        """
        return chain_factory(file)

    def to_file(self):
        """ 通过chain对象，生成chain肖像文件
        """
        # todo: 待完成编码
        pass



import hashlib

HMAC_BLOCK_SIZE = 64


def hash256(m):
    if type(m) != bytes:
        m = m.encode("utf-8")
    return hashlib.sha256(m).digest()


def hash512(m):
    if type(m) != bytes:
        m = m.encode("utf-8")
    return hash256(m + bytes([0])) + hash256(m + bytes([1]))


def hmac256(m, k):
    if type(m) != bytes and type(m) != bytearray:
        m = m.encode("utf-8")
    if type(k) != bytes and type(k) != bytearray:
        k = k.encode("utf-8")
    k = bytes(k)
    if len(k) > HMAC_BLOCK_SIZE:
        k = hash256(k)
    while len(k) < HMAC_BLOCK_SIZE:
        k += bytes([0])
    opad = bytes([0x5C] * HMAC_BLOCK_SIZE)
    ipad = bytes([0x36] * HMAC_BLOCK_SIZE)
    kopad = bytes([k[i] ^ opad[i] for i in range(HMAC_BLOCK_SIZE)])
    kipad = bytes([k[i] ^ ipad[i] for i in range(HMAC_BLOCK_SIZE)])
    return hash256(kopad + hash256(kipad + m))


"""
Copyright 2020 Chia Network Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



import os
from typing import List

from fabric import Connection, Config
from platon_env.base.supervisor.supervisor import Supervisor
from platon_env.utils.md5 import md5
from platon_env.utils.path import join_path
from platon_env.base.process import Process


class Host:
    supervisor: Supervisor = None
    processes: dict = dict()

    def __init__(self,
                 ip: str,
                 username: str,
                 certificate: str = None,
                 password: str = None,
                 port: int = 22,
                 is_superviosr: bool = True,
                 processes: List[Process] = None
                 ):
        """
        初始化远程主机对象并尝试连接，支持免密/密码/证书方式。

        Args:
            ip: IP地址
            username: 用户名
            certificate: 证书
            password: 密码/证书密码
            port: ssh端口
            processes: 要注册的进程
        """
        self.ip = ip
        self.port = port
        self.username = username
        self.certificate = certificate
        self.password = password
        self._connection = None
        self.home_dir = join_path('/home', self.username)
        self.tmp_dir = '.env-tmp'
        self.is_superviosr = is_superviosr
        if self.is_superviosr:
            self.supervisor = Supervisor(self)
            self.prepare()
        if not processes:
            processes = []
        for process in processes:
            self.register(process)

    def __eq__(self, other):
        if self.ip == other.ip and self.username == other.username:
            return True
        return False

    def __hash__(self):
        return hash(id(self))

    @property
    def connection(self):
        """ 单例模式连接到服务器,支持免密/密码/证书方式
        """
        if self._connection and self._connection.is_connected:
            return self._connection

        if self.certificate:
            # todo: 证书方式连接
            config = None
        else:
            config = Config(overrides={'sudo': {'password': self.password}})

        self._connection = Connection(self.ip, self.username, self.port, config=config, connect_kwargs={'password': self.password})
        return self._connection

    def prepare(self):
        """ 主机环境准备
        """
        self.supervisor.install()

    def pid(self, name):
        """ 通过进程名字，获取远程主机的进程号
        """
        return self.ssh(f'ps -ef | grep {name} | grep -v grep | ' + "awk {'print $2'}")

    def ssh(self, command, sudo=False, warn=True, hide='both', strip=True):
        """ 在远程主机上执行ssh命令

        Args:
            command: 需要执行的shell命令行
            sudo: 是否用sudo的方式执行
            warn: 当命令执行失败的时候，以警告的方式打印错误，而不是抛出异常
            hide: 打印错误信息时，需要隐藏的输出内容，包含：out、err、both
            strip: 是否直接返回命令行执行结果，False返回fabric-result对象
        """
        if sudo:
            result = self.connection.sudo(command, warn=warn, hide=hide)
        else:
            result = self.connection.run(command, warn=warn, hide=hide)
        if strip:
            return result.stdout.strip()
        return result

    def file_exist(self, path):
        """ 判断文件是否已存在于远程主机
        """
        return self.ssh(f'test -e {path}', strip=False).ok

    def fast_put(self, local, remote=None, sudo=False):
        """ 使用缓存机制，上传文件到远程主机，以提高上传速度
        # 注意：并发推送时，出现过低概率put失败问题，请慎用
        # todo: 支持压缩上传
        # todo: 检查文件是否已经缓存的时候，增加判断文件大小
        """
        _md5 = md5(local)
        tmp_file = join_path(self.tmp_dir, _md5)
        if not self.file_exist(self.tmp_dir):
            self.ssh(f'mkdir -p {self.tmp_dir}')
        if not self.file_exist(tmp_file):
            try:
                # todo: 并发推送genesis.json时，在该步骤fabric概率报错，原因尚未找到
                self.connection.put(local, tmp_file)
            except Exception as e:
                # todo: 按异常类型处理
                raise Exception(f'fast put {local} to {remote} error: {e}')
        if not remote:
            return tmp_file
        path, _ = os.path.split(remote)
        if not self.file_exist(path):
            self.ssh(f'mkdir -p {path}', sudo=sudo)
        self.ssh(f'cp {tmp_file} {remote}', sudo=sudo)

    def write_file(self, content: str, file):
        """ 将文本内容写入远程主机的文件，目前仅支持写入新的文件
        # todo： 支持写入已存在的文件，包括覆盖、追加等方式
        """
        path, _ = os.path.split(file)
        if path:
            self.ssh(f'mkdir -p {path}')

        content = str(content).replace('"', r'\"')
        self.ssh(rf'echo "{content}" > {file}')

    def register(self, process: Process):
        """ 将进程注册到主机对象，以便后续管理和使用
        """
        if self.processes.get(process.name):
            raise Exception()

        self.processes[process.name] = process

    def unregister(self, name):
        """ 通过进程名字，删除远程主机上的进程注册信息
        """
        self.processes.pop(name)

    def get_processes(self, **kwargs):
        """ 通过进程属性，获取进程对象，可以同时匹配多个
        """
        processes = []

        def match(process: Process):
            for key, value in kwargs:
                if process.__getattribute__(key) == value:
                    return process

        for process in self.processes:
            matched = match(process)
            if matched:
                processes.append(matched)

        return processes

    def set_supervisor(self, supervisor: Supervisor):
        self.supervisor = supervisor



from typing import List

from platon_env.base.process import Process


class Service:
    processes: dict = dict()

    def __init__(self, processes: List[Process] = None):
        if not processes:
            processes = []
        for process in processes:
            self.add_process(process)

    def add_process(self, process: Process):
        """ 将进程添加到服务，进行统一管理
        """
        if self.processes.get(id(process)):
            raise Exception('The process already exists.')

        self.processes[id(process)] = process

    def remove_process(self, name):
        """ 从服务中移除进程
        """
        self.processes.pop(name)

    def get_processes(self, **kwargs):
        """ 通过进程属性，获取进程对象，可以同时匹配多个
        """
        processes = []

        def match(process: Process):
            for key, value in kwargs:
                if process.__getattribute__(key) == value:
                    return process

        for process in self.processes:
            matched = match(process)
            if matched:
                processes.append(matched)

        return processes

    # def show(self, *args, **kwargs):
    #     raise NotImplementedError("process must implement this method")

    def install(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def uninstall(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def status(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def start(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def restart(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def stop(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")



import json
from typing import List

from platon_env.node import Node


class Genesis:

    def __init__(self, basic_file: str = None):
        self._basic_file = basic_file
        with open(basic_file, mode='r', encoding='utf-8') as file:
            self.data = json.load(file)

    @property
    def init_node(self):
        return self.data['config']['cbft']['initialNodes']

    def fill_init_nodes(self, nodes: List[Node] = None, content: List[dict] = None):
        """ 填写初始验证人信息，支持传入node对象列表，或者初始验证人信息
        """
        init_nodes = []
        if nodes:
            for node in nodes:
                init_nodes.append({"node": node.enode, "blsPubKey": node.bls_pubkey})
        elif content:
            init_nodes = content
        self.data['config']['cbft']['initialNodes'] = init_nodes
        return self

    def fill_accounts(self, accounts: dict):
        """ 填写初始账户信息
        """
        self.data['alloc'] = accounts
        return self

    def save(self, file):
        """ 保存为本地文件
        """
        with open(file, mode='w', encoding='utf-8') as f:
            f.write(json.dumps(self.data, indent=4))



import json
from dataclasses import dataclass, field, asdict
from typing import List
from dacite import from_dict
from ruamel import yaml

from platon_env.base.host import Host
from platon_env.node import Node
from platon_env.utils.key.keytool import gen_node_keypair, gen_bls_keypair


@dataclass
class CommonData:
    host: str = None
    username: str = None
    password: str = None
    ssh_port: int = 22
    p2p_port: int = None
    options: str = None

    def _fill_common_info(self, node):
        info_dict = asdict(self)
        for k, v in info_dict.items():
            if not v:
                continue
            if hasattr(node, k) and not getattr(node, k):
                setattr(node, k, v)

    def to_dict(self):
        return asdict(self)


@dataclass
class NodeData(CommonData):
    node_id: str = None
    node_key: str = None
    bls_pubkey: str = None
    bls_prikey: str = None

    def __post_init__(self):
        self._fill_node_info()

    def _fill_node_info(self):
        if not self.node_id or not self.node_key:
            node_id, node_key = gen_node_keypair()
            self.node_id = node_id
            self.node_key = node_key
        if not self.bls_pubkey or not self.bls_prikey:
            bls_pubkey, bls_prikey = gen_bls_keypair()
            self.bls_pubkey = bls_pubkey
            self.bls_prikey = bls_prikey


@dataclass
class NodeGroupData(CommonData):
    nodes: List[NodeData] = field(default_factory=[])

    def __post_init__(self):
        for node in self.nodes:
            self._fill_common_info(node)


@dataclass
class ChainData(CommonData):
    deploy_dir: str = None
    init: NodeGroupData = None
    normal: NodeGroupData = None

    def __post_init__(self):
        init_nodes = self.init.nodes if self.init else []
        normal_nodes = self.normal.nodes if self.normal else []
        nodes = init_nodes + normal_nodes

        for node in nodes:
            self._fill_common_info(node)
        self.__check_nodes()

    def __check_nodes(self):
        """ 检查节点信息是否完善与正确
        """
        # todo: 完成编码
        pass


def _create_dataclass(cls, _dict):
    """ 将dict数据转换为dataclass对象
    """
    return from_dict(cls, _dict)


def _save_dataclass(obj, file):
    """ 将dataclass对象存储为文件
    # todo: 实现存储为yaml文件
    """
    data = obj.to_dict()
    with open(file, "w") as f:
        f.write(json.dumps(data, indent=4))


def chain_factory(file: str):
    """ 根据chain配置文件生成chain对象
    # todo: 支持无密码连接
    """
    with open(file, encoding='utf-8') as f:
        data = yaml.load(f)
    chain_data = _create_dataclass(ChainData, data['chain'])

    hosts, nodes = [], []

    # todo: 优化该方法
    init_nodes = chain_data.init.nodes if chain_data.init else []
    for member in init_nodes:
        host = Host(member.host,
                    member.username,
                    password=member.password,
                    port=member.ssh_port,
                    is_superviosr=True,
                    )
        hosts.append(host)
        node = Node(host,
                    member.node_id,
                    member.node_key,
                    p2p_port=member.p2p_port,
                    bls_pubkey=member.bls_pubkey,
                    bls_prikey=member.bls_prikey,
                    options=member.options,
                    is_init_node=True,
                    base_dir=chain_data.deploy_dir,
                    )
        nodes.append(node)

    normal_nodes = chain_data.normal.nodes if chain_data.normal else []
    for member in normal_nodes:
        host = Host(member.host,
                    member.username,
                    password=member.password,
                    port=member.ssh_port,
                    is_superviosr=True,
                    )
        hosts.append(host)
        node = Node(host,
                    member.node_id,
                    member.node_key,
                    p2p_port=member.p2p_port,
                    bls_pubkey=member.bls_pubkey,
                    bls_prikey=member.bls_prikey,
                    options=member.options,
                    is_init_node=False,
                    base_dir=chain_data.deploy_dir,
                    )
        nodes.append(node)

    from platon_env.chain import Chain
    return Chain(nodes)



import os

from loguru import logger

from platon_env.base.process import Process
from platon_env.utils.path import join_path


class Supervisor(Process):
    base_path = '/etc/supervisor'
    config_file = '/etc/supervisor/supervisord.conf'
    process_config_path = '/etc/supervisor/conf.d'
    status_command = 'supervisorctl status {}'
    start_command = 'supervisorctl start {}'
    restart_command = 'supervisorctl restart {}'
    stop_command = 'supervisorctl stop {}'
    update_command = 'supervisorctl update {}'

    def install(self):
        """ 安装supervisor
        """
        is_installed = self.host.ssh("apt list | grep supervisor", warn=False)
        if '[installed]' not in str(is_installed):
            self.host.ssh('apt update', sudo=True, warn=False)
            self.host.ssh('apt install -y --reinstall supervisor', sudo=True, warn=False)
            self._upload_config()
            logger.info(f'Supervisor install success!')

        pid = self.host.pid('supervisord')
        if not pid:
            # todo: 增加 config_file 不存在的判断
            self.host.ssh(f'supervisord -c {self.config_file}', sudo=True, warn=False)
            self.host.ssh('chmod 770 /var/run/supervisor.sock', sudo=True)

    def uninstall(self):
        """ 卸载supervisor
        """
        self.host.ssh(f'apt remove supervisor', sudo=True, warn=False)
        logger.info(f'Supervisor uninstall success!')

    def add(self, name, config=None, file=None):
        """ 为supervisor添加要管理的进程，支持name + config的方式，或者直接上传配置文件以添加进程。
        """
        config_path = join_path(self.process_config_path, f'{name}.conf')

        if config:
            self.host.ssh(f"sh -c 'echo \"{config}\" > {config_path}'", sudo=True, warn=False)
        elif file:
            self.host.fast_put(file, config_path)
        else:
            raise ValueError('All parameters are not available.')
        self.update(name)

    def remove(self, name):
        """ 通过进程名称，删除supervisor管理的进程
        """
        process_file = join_path(self.process_config_path, f'{name}.conf')
        self.host.ssh(f'rm -rf {process_file}', sudo=True, warn=False)
        self.update()

    def status(self, name='') -> bool:
        """ 通过进程名称，查看进程状态
        """
        outs = self.host.ssh(self.status_command.format(name))
        if not outs or 'RUNNING' in str(outs):
            return True
        return False

    def update(self, name=''):
        """ 更新supervisor的进程列表
        """
        # self.host.connection(self.update_command)
        self.host.ssh(self.update_command.format(name), warn=False)

    def start(self, name):
        """ 启动进程
        """
        self.host.ssh(self.start_command.format(name), warn=False)

    def restart(self, name):
        """ 重启进程
        """
        self.host.ssh(self.restart_command.format(name), warn=False)

    def stop(self, name):
        """ 停止进程
        """
        self.host.ssh(self.stop_command.format(name), warn=False)

    def _upload_config(self, file=None):
        """ 上传supervisor的配置文件
        """
        if not file:
            current_path = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
            file = join_path(current_path, 'supervisor.conf')
        self.host.fast_put(file, self.config_file)



import os
import re
import tarfile
from typing import Union, List, Literal

from loguru import logger
from paramiko.ssh_exception import SSHException

from platon_env.base.host import Host
from platon_env.base.process import Process
from platon_env.utils.path import join_path


class NodeOpts:

    def __init__(self,
                 rpc_port: int = None,
                 rpc_api: str = None,
                 ws_port: int = None,
                 ws_api: str = None,
                 extra_opts: str = None
                 ):
        self.rpc_port = rpc_port
        self.rpc_api = rpc_api
        self.ws_port = ws_port
        self.ws_api = ws_api
        self.extra_opts = extra_opts

    def __str__(self):
        string = ''
        if self.rpc_port:
            assert self.rpc_api, 'The RPC API is not defined.'
            string += f'--rpc --rpcaddr 0.0.0.0 --rpcport {self.rpc_port} --rpcapi {self.rpc_api} '
        if self.ws_port:
            assert self.ws_api, 'The WS API is not defined.'
            string += f'--ws --wsaddr 0.0.0.0 --wsport {self.rpc_port} --wsapi {self.rpc_api} '
        if self.extra_opts:
            string += self.extra_opts
        return string

    def to_string(self):
        return str(self)


default_opts = NodeOpts(rpc_port=7789,
                        rpc_api='platon,txpool,net,debug',
                        ws_port=7790,
                        ws_api='platon,txpool,net,debug'
                        )


class Node(Process):

    def __init__(self,
                 host: Host,
                 node_id: str,
                 node_key: str,
                 p2p_port: int = 16789,
                 bls_pubkey: str = None,
                 bls_prikey: str = None,
                 options: Union[str, NodeOpts] = '',
                 is_init_node: bool = False,
                 base_dir: str = 'platon',
                 ):
        super().__init__(host, base_dir=base_dir, port=p2p_port)
        self.node_id = node_id
        self.node_key = node_key
        self.p2p_port = p2p_port
        self.bls_pubkey = bls_pubkey
        self.bls_prikey = bls_prikey
        self.options = options
        self.is_init_node = is_init_node
        # 连接信息
        self.static_nodes = None
        # 部署信息
        self.name = f'p{self.p2p_port}'
        self.base_dir = base_dir
        if not os.path.isabs(self.base_dir):
            self.base_dir = join_path(self.host.home_dir, self.base_dir)
        self.deploy_path = join_path(self.base_dir, self.name)
        self.platon = join_path(self.deploy_path, 'platon')
        self.genesis_file = join_path(self.deploy_path, 'genesis.json')
        self.node_key_file = join_path(self.deploy_path, 'nodekey')
        self.bls_prikey_file = join_path(self.deploy_path, 'blskey')
        self.data_dir = join_path(self.deploy_path, 'data')
        self.static_file = join_path(self.data_dir, 'static-nodes.json')
        self.keystore_dir = join_path(self.data_dir, 'keystore')
        self.log_dir = join_path(self.deploy_path, 'log')
        self.log_file = join_path(self.log_dir, 'platon.log')
        self.supervisor_file = join_path(self.host.supervisor.process_config_path, self.name + '.conf')

    def __str__(self):
        return f'{self.host.ip}:{self.p2p_port}'

    def __eq__(self, other):
        if self.host == other.host and self.p2p_port == other.p2p_port:
            return True
        return False

    def __hash__(self):
        return hash(id(self))

    @property
    def enode(self):
        """ 获取节点的enode信息
        """
        return f"enode://{self.node_id}@{self.host.ip}:{self.p2p_port}"

    def gql(self, scheme: Literal['ws', 'wss', 'http', 'https'] = 'http'):
        """ 获取节点的graphql连接信息
        注意：当前仅支持http方式，其他scheme为超前设计
        """
        match = re.search('--graphql', self.options)

        if match:
            return f"{self.rpc(scheme)}/graphql/platon"

        raise ValueError(f'The {scheme} graphql is not open.')

    def rpc(self, scheme: Literal['ws', 'wss', 'http', 'https', 'ipc'] = 'ws'):
        """ 获取节点的rpc连接信息
        """
        options = self.options + ' '  # 在后面添加' '，避免出现miss match
        ws_match = re.search('--wsport (.+?) ', options)
        http_match = re.search('--rpcport (.+?) ', options)
        ipc_match = re.search('--ipcpath (._?) ', options)

        if (scheme == 'ws' or scheme == 'wss') and ws_match:
            return f"{scheme}://{self.host.ip}:{ws_match.group(1)}"

        if (scheme == 'http' or scheme == 'https') and http_match:
            return f"{scheme}://{self.host.ip}:{http_match.group(1)}"

        if scheme == 'ipc' and ipc_match:
            return ipc_match.group(1)

        raise ValueError(f'The {scheme} rpc is not open.')

    def install(self,
                platon: str,
                network: str,
                genesis_file: str = None,
                static_nodes: List[str] = None,
                keystore: str = None,
                options: str = '',
                ):
        """ 使用supervisor部署节点
        """
        if not self.host.supervisor:
            raise Exception("supervisor not install.")
        self.uninstall()

        # 准备部署所需的文件
        self.upload_platon(platon)
        self.host.write_file(self.node_key, self.node_key_file)
        if network == 'private':
            if not genesis_file:
                raise ValueError('Private network needs genesis file.')
            self.host.connection.put(genesis_file, self.genesis_file)
        if self.bls_prikey:
            self.host.write_file(self.bls_prikey, self.bls_prikey_file)

        if static_nodes:
            self.set_static_nodes(static_nodes)
        if keystore:
            self.upload_keystore(keystore)

        # 启动节点
        self.init()
        self.start(options)
        logger.info(f'Node {self} install success!')

    def uninstall(self):
        """ 清理节点，会停止节点并删除节点文件
        """
        self.stop()
        self.host.ssh(f'rm -rf {self.deploy_path}', sudo=True)
        self.host.supervisor.remove(self.name)
        logger.info(f'Node {self} uninstall success!')

    def status(self) -> bool:
        """ 获取节点的运行状态
        """
        return self.host.supervisor.status(self.name)

    def init(self):
        """ 初始化节点
        """
        self.host.ssh(f'mkdir -p {self.log_dir}')
        result = self.host.ssh(f'{self.platon} --datadir {self.data_dir} init {self.genesis_file}',
                               warn=False, strip=False)
        if result.failed or ('Fatal' in result.stderr or 'Error' in result.stderr):
            raise SSHException(result.stderr)
        logger.info(f'Node {self} init success!')

    def start(self, options: Union[str, NodeOpts] = ''):
        """ 使用supervisor启动节点
        """
        self.options = str(options) or str(self.options)
        self.host.supervisor.add(self.name, self.supervisor_config)
        self.host.supervisor.start(self.name)
        logger.info(f'Node {self} start success!')

    def restart(self):
        """ 使用supervisor重启节点
        """
        self.host.supervisor.restart(self.name)
        logger.info(f'Node {self} restart success!')

    def stop(self):
        """ 使用supervisor停止节点
        """
        self.host.supervisor.stop(self.name)
        logger.info(f'Node {self} stop success!')

    def upload_platon(self, platon_file):
        """ 使用缓存上传platon
        """
        self.host.fast_put(platon_file, self.platon)
        self.host.ssh(f'chmod u+x {self.platon}')
        logger.debug(f'Node {self} upload platon success!')

    def upload_keystore(self, keystore: str):
        """ 使用缓存上传keystore，支持单个文件与目录
        """
        if os.path.isfile(keystore):
            self.host.fast_put(keystore, self.keystore_dir)
        elif os.path.isdir(keystore):
            tar_file = keystore + '.tar.gz'
            tar = tarfile.open(tar_file, "w:gz")
            tar.add(keystore, arcname=os.path.basename(keystore))
            tar.close()
            tmp_file_path = self.host.fast_put(tar_file)
            self.host.ssh(f'tar xzvf {tmp_file_path} {self.keystore_dir}')
        else:
            raise FileNotFoundError('keystore not found.')
        logger.debug(f'Node {self} upload keystore success!')

    def set_static_nodes(self, enodes: List[str]):
        """ 指定要连接的静态节点，可以指定多个
        """
        for enode in enodes:
            assert 'enode://' in enode, 'enode format is incorrect.'
        self.static_nodes = enodes

        formatted_enodes = str(enodes).replace("'", '"')
        self.host.write_file(formatted_enodes, self.static_file)
        logger.debug(f'Node {self} set static nodes success!')

    @property
    def supervisor_config(self):
        """ 在远程主机生成节点的supervisor配置文件
        """
        program = f'[program:{self.name}]\n'
        bls_key_cmd = f'--cbft.blskey {self.bls_prikey_file} ' if self.bls_prikey else ''
        command = (f'command={self.platon} --identity {self.name} '
                   f'--datadir {self.data_dir} '
                   f'--port {self.p2p_port} '
                   f'--nodekey {self.node_key_file} '
                   f'{bls_key_cmd}'
                   f'{self.options} \n')
        setting = '\n'.join(['numprocs=1', 'autostart=false', 'startsecs=1', 'startretries=3', 'autorestart=unexpected',
                             'exitcode=0', 'stopsignal=TERM', 'stopwaitsecs=10', 'redirect_stderr=true',
                             'stdout_logfile_maxbytes=200MB', 'stdout_logfile_backups=20',
                             f'stdout_logfile={self.log_file}'])
        config = program + command + setting
        return config



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from platon_env.base.host import Host


class Process:

    def __init__(self, host: 'Host', name=None, base_dir=None, port=None, pid=None):
        self.host = host
        self.name = name
        self.base_dir = base_dir
        self.port = port
        self.pid = pid
        # if self.base_dir:
        #     self.host.ssh(f'mkdir -p {self.base_dir}',sudo=False)

    def check(self, *args, **kwargs):
        pass

    def install(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def uninstall(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def status(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def start(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def restart(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")

    def stop(self, *args, **kwargs):
        raise NotImplementedError("process must implement this method")



import os
import sys

from platon_keys import keys
from platon_utils.curried import keccak, text_if_str, to_bytes

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def gen_node_keypair(extra_entropy=''):
    extra_key_bytes = text_if_str(to_bytes, extra_entropy)
    key_bytes = keccak(os.urandom(32) + extra_key_bytes)
    private_key = keys.PrivateKey(key_bytes)
    return keys.private_key_to_public_key(private_key).to_hex()[2:], private_key.to_hex()[2:],


def gen_bls_keypair():
    if 'linux' in sys.platform:
        tool_file = os.path.abspath(os.path.join(BASE_DIR, 'bin/keytool'))
        execute_cmd('chmod +x {}'.format(tool_file))
    elif 'win' in sys.platform:
        tool_file = os.path.abspath(os.path.join(BASE_DIR, 'bin/keytool.exe'))
    else:
        raise Exception('This platform is not supported currently')
    keypair = execute_cmd(f'{tool_file} genblskeypair')
    if not keypair:
        raise Exception('Unable generate bls keypair')
    lines = keypair.split('\n')
    private_key = lines[0].split(':')[1].strip()
    public_key = lines[1].split(':')[1].strip()
    if not private_key or not public_key:
        raise Exception('Incorrect bls keypair')
    return public_key, private_key


def execute_cmd(cmd):
    r = os.popen(cmd)
    out = r.read()
    r.close()
    return out



from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED
from functools import wraps

error_msg = '{} {} do {} failed:{}'


def executor_wrapper(obj, func_name):
    """ 在try中执行方法，当报错时返回错误信息

    Args:
        obj: 需要执行方法的节点列表
        func_name (str): 需要执行的方法名
    """
    func = obj.__getattribute__(func_name)

    @wraps(func)
    def wrap_func(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            return False, error_msg.format(type(obj), obj, func.__name__, e)
        # todo: 正常返回时，增加对应的obj信息
        return True, result

    return wrap_func


def concurrent_executor(objects, func_name, *args, **kwargs):
    """ 使用线程池并行执行方法

    Args:
        objects: 需要执行方法的对象列表
        func_name (str): 需要执行的方法名
        args: 方法参数
        kwargs: 关键字参数
    Returns:
        succeed: 正常执行成功后的返回信息
    """
    succeed, failed = [], []
    with ThreadPoolExecutor(max_workers=20) as thread_pool:
        runs = []
        for obj in objects:
            func = executor_wrapper(obj, func_name)
            runs.append(thread_pool.submit(func, *args, **kwargs))
        done, undone = wait(runs, timeout=30, return_when=ALL_COMPLETED)
    for piece in done:
        is_succeed, result = piece.result()
        succeed.append(result) if is_succeed else failed.append(result)
    if failed:
        raise Exception(f'executor {func.__name__} failed: {failed}')
    # todo: 增加对undone的处理
    return succeed
